// printer.cpp : Printing via hprdh.dll
//

#include "stdafx.h"
#include <windows.h>
#include <winspool.h>
#include <winioctl.h>
#include <winuser.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "HPRDH.h"
#include "resource.h"

#define	CLAIM_TIMEOUT				20000
#define NUMBER_OF_RETRIES			5
#define MAX_WAIT_FOR_READ			5000	// Wait time for next read in milli sec
#define MAX_WAIT_FOR_EJECT			60000	// Wait time for eject
#define WAIT_TIME_RETRY				500		// Wait time for retry in milli seconds
									// Wait time for read thread in milli seconds
#define WAIT_TIME_READ_THREAD		500
#define WAIT_TIME_READ				1000	// Wait time read in milli seconds
#define	TYPE_SIZE					16
#define MAX_WRITE_STRING			128
#define NUMBER_OF_LINES				10
#define NUMBER_OF_PAGES				2

#define	ERROR_STRING_SIZE			1024

								// Eject paper, will be ignored, if there is none
#define EJECT			 "\033[0$p"	
#define SETCTLP			 "\033[%d!t"	// Set control point 1
#define SETCTLP2		 "\033[%d)t"	// Set control point 2
#define INIT_PRINTER	 "\033@"		// initialize the printer
#define GRAPH_CHAR_SET	 "\033t\001"	// Select graphical character set
#define EXT_CHAR_SET	 "\0336"		// Switching on extended character set
#define US_CHAR_SET		 "\033R\000"	// Selecting the character set of the USA
#define LEFT_MARGIN10	 "\033l\012"	// Setting the left margin to 10
#define LQ_PRINT_QUALITY "\033x\001"	// Switch to LQ print quality
#define	INCH_LINE_SPACE	 "\0332"		// Selecting 1/6 inch line spacing
								// Determing the document length in lines to 60 
#define	DOC_LEN_IN_LINES_60 "\033C\074"
#define	VERT_POS_100	 "\033[100!p"	// Absolute vertical pos. from the top (100)
#define	HORIZ_POS		 "\033$\030\0"	// Absolute horizontal positioning 

#define PRINTING_UNIT_PARAMETERS	"\033[1#z" 	

#define	SPECIAL_PRINTER_STATE "\033[\"z"	// Get special printer state
#define FF				"\014"				// Sheet ejection

#define GLOB_UNDEF	    -1
#define GLOB_ONLINE     0
#define GLOB_STOP		1 
#define GLOB_ERROR      2

#define SPC_PAPER_STATE 4
#define SPC_PAPER		1
#define SPC_NOPAPER		2

#define	SPC_USER_KEY	20
#define	SPC_USER_KEY1	1
#define	SPC_USER_KEY2	2


	// Handle to wake up the read thread, if information from the printer is wanted
extern HANDLE	hWakeupThread;			// or if the thread has to end		
	// Handle to sequence access to the common data of main and read thread
extern HANDLE	hReadData;
	// Handle to wake up the main thread from the read thread, if information
extern HANDLE	hWakeUp;		// from the printer is available
	// The main thread sets StopThread to TRUE, if the read thread has to end
extern BOOL		StopThread;
extern DWORD	DlgResult;	// Handle to the dialog box, which asks to insert paper
	// Contains the information, if the user has pressed the cancel button
extern HWND		DlgHwnd;		// of the paper dialog

		// Function to print the desired message in a dialog box
extern void	PrintMessage(BOOL bNewThread, char *PtrOutMessage, DWORD dwError, char *PtrMessage, ...);
// Create dialog thread for to enter paper or to inform about printer stopped
extern BOOL CreateDialogThread(char *pcText);	

int WriteToPrinter(char *pcWriteString, LPDWORD pdwNoOfBytesWritten, 
				   DWORD dwRetAllowed, DWORD dwRetAllowed2, BOOL CheckPaper);

// The information returned from PrtRead() is put into this structure.
// If a control point is reached (PPRT_CONTROL_POINT or PRT_CONTROL_POINT_2)
// or the global state (PRT_GLOBAL_STATUS) or a special message 
// (PRT_SPECIAL_MESSAGE) from the printer is returned, this information
// is not returned in such a structure. In that case global variables 
// (lControlPointReached, lControlPoint2Reached, lGlobalStatus,
//  lSpecialStatus1 and lSpecialStatus2 contain the desired information
typedef struct
{
	  					// The type of information got from the printer.
      DWORD  dwType;	// (See the possible types in HPRDH.h)
							   // The part after ESC [ always contains numerical 
							   // parameter in the following syntactic structure:
							   //		P1; ... ;Pn
							   // That is why we offer these parameter for the 
							   // PrtRead() in a numeric array.
							   // The numeric parameter are saved here numerically.	
	  DWORD	*pdwParArray;  
      DWORD  dwNoOfElements;   // The number of numerical parameters in dwParArray. 
      CHAR   *pcReadData;	   // The data received between STX and ETX.
      DWORD  dwReadDataLen;    // Size of the contents of pcReadData
							   // Pointer on the next entry in the read queue.
	  VOID   *pvNextReadQueue; // The pointer is NULL, if there is no next entry.

} ReadQueue;

								// Pointer on the first entry in the ReadQueue.
ReadQueue  *pvStartReadQueue;	// If the ReadQueue is empty this pointer is NULL.
ReadQueue  *pvEndReadQueue;		// Pointer on the last entry in the ReadQueue

	// Handle to the printer, which is returned from PrtOpen() 
	// and has to be given to PrtClaim(), PrtRead(), PrtWrite() PrtRelease()
HANDLE hPrinter = NULL;		// and PrtClose()
					// Contains the value of the control point wanted 
LONG  lControlPointWanted = -1;
					// Contains the value of the control point 2 wanted
LONG  lControlPoint2Wanted = -1;
					// Contains the last global status returned from the printer
					// Contains the value of the last control point reached 
LONG  lControlPointReached = -1;
					// Contains the value of the last control point 2 reached
LONG  lControlPoint2Reached = -1;
					// Contains the last global status returned from the printer
LONG  lGlobalStatus = GLOB_UNDEF;
		// The following two variables contain the last special status returned
LONG  lSpecialStatus1 = -1;
LONG  lSpecialStatus2 = -1;

DWORD dwRetLastRead = PRT_NO_ERROR;
DWORD dwReadWaitFlag = 0;

/* Definition of the ReadWait flag */

#define NO_PAPER			  001 // Bit 0: Wait until paper is ejected
#define PAPER				  002 // Bit 1: Wait until paper is entered
#define USER1_KEY			  004 // Bit 2: Wait until USER 1 key is pressed
#define USER2_KEY			  010 // Bit 3: Wait until USER 2 key is pressed
#define GLOBAL_STATE		  020 // Bit 4: Wait for a new global state
#define CONTROL_POINT		  040 // Bit 5: Wait until control point is reached
#define CONTROL_POINT_2 	 0100 // Bit 6: Wait until control point 2 is reached
#define DOCUMENT_WIDTH		 0200 // Bit 7: Wait until document width is returned
#define MSR_MICR_DATA	     0400 // Bit 8: Wait for the read data of MSR track/MICR
#define MSR_WRITE_STATE	    01000 // Bit 9: Wait for the state of MSR write
#define PRINTING_UNIT_PARAM 02000 // Bit 10: Wait for printing unit parameters
#define MOD_ID				04000 // Bit 11: Wait for the data id of the module
		

// Contains, what has to be done, if printing has to be stopped after PrtClaim()
// has to be called successfully.
void EndPrinting() {
	// Wait until no other thread has access control on the common data
	WaitForSingleObject(hReadData, INFINITE);
	StopThread = TRUE;			// Tell the read thread, that he has to stop
	SetEvent(hWakeupThread);	// Wake up the read thread, if it is waiting
	SetEvent(hReadData);		// Allow other threads again to access common data

	PrtRelease(hPrinter);		// Inform, that printing of this job has finished
	PrtClose(&hPrinter);		// Close connection to the printer
	return;
}

/*++

Routine Description:
    Delete this entry of the read queue

Arguments:
    pReadQueueEntry: Read queue entry, which has to be removed.
		
Return Value:
	None
--*/

void DeleteReadQueueEntry(ReadQueue *pReadQueueEntry) {
	char		*pcReadData;
	DWORD		*pdwParArray;
	
	pcReadData = (char *) pReadQueueEntry->pcReadData;

	if (pcReadData != NULL) {
		// If this entry contains information in pcReadData
		// free this allocated space
		free((void *) pcReadData);
	}

	pdwParArray = pReadQueueEntry->pdwParArray;
	if (pdwParArray != NULL) {
		// If this entry contains an array of variables
		// free the corresponding allocated space
		free((void *) pdwParArray);
	}
	
	// Free the allocated space of this readqueue entry
	free((void *) pReadQueueEntry);
	return;
}

/*++

Routine Description:
    Remove the first entry out of the read queue
	(The read queue entry is only removed out of the queue,
	 but not removed itself!)

Arguments:
    	
Return Value:
	None
--*/

void RemoveFirstReadQueueEntry() {
	ReadQueue	*pFirstReadQueue;
	
	pFirstReadQueue = pvStartReadQueue;
	if (pFirstReadQueue == NULL) {
		return;		// The whole read queue is already empty
	}
	// Get the next entry in the read queue, because this is now
	// the first entry in the read queue.
	pvStartReadQueue = (ReadQueue *) pFirstReadQueue->pvNextReadQueue;

	if (pvStartReadQueue == NULL) {
		// The read is empty, after the first entry is removed.
		// That is why the pointer on the last entry has to be set to NULL.
		pvEndReadQueue = NULL;
	}

	return;
}


/*++

Routine Description:
    Delete the first entry of the read queue
	(Here the first entry is also removed itself)

Arguments:
    	
Return Value:
	None
--*/

void DeleteFirstReadQueueEntry() {
	ReadQueue	*pFirstReadQueue;
	
	pFirstReadQueue = pvStartReadQueue;
	if (pFirstReadQueue != NULL) {
		RemoveFirstReadQueueEntry();
		DeleteReadQueueEntry(pFirstReadQueue);
	}
	return;
}

/*++

Routine Description:
    Delete all entries of the read queue
	
Arguments:
    	
Return Value:
	None
--*/
void ClearReadQueue() {
	// Wait until no other thread has access control on the common data
	WaitForSingleObject(hReadData, INFINITE);
	while (pvStartReadQueue != NULL) {
		DeleteFirstReadQueueEntry();
	}
	SetEvent(hReadData);		// Allow other threads again to access common data
}

/*++

Routine Description:
    Wake up read thread.
	The access to the common data must be guaranteed 
	(WaitForSingleObject(hReadData, INFINITE);)
	If the read thread ended because of a read error,
	the main thread exits also.
	
Arguments:
    	
Return Value:
	None
--*/
void WakeUpReadThread() {
					// Did the read thread end, because of an error in PrtRead() ?
	if (dwRetLastRead != PRT_NO_ERROR) {
																		  // yes
		SetEvent(hReadData);
		PrintMessage(FALSE, NULL, dwRetLastRead, "Read Printer");
		EndPrinting();
		exit(3);
	}
						// wake up the thread to do the next read, because we wait
	SetEvent(hWakeupThread); // for the information, that paper is entered
				// Init handle hWakeUp, that this thread waits during next call of 
	ResetEvent(hWakeUp);				// WaitForSingleObject(hWakeUp)
	return;
}

/*++

Routine Description:
    Get the desired entry of the read queue.
	If the entry is not yet in the queue,
	wake up the read thread to read again.
	All entries in the read queue before this entry are removed out of the
	queue and are deleted.
	The desired entry is removed out of the queue, but not deleted.
	The calling thread has to delete this entry later itself.
	
Arguments:
	dwCurrentType: The type of the wanted read queue entry
	pReadQueueEntry: Return the pointer on the desired read queue entry here
    	
Return Value:
	None
--*/

DWORD GetInformationOfDesiredType(DWORD dwCurrentType, ReadQueue **pReadQueueEntry) { 
	DWORD	dwType = -1;
	DWORD	dwRet;

	*pReadQueueEntry = NULL;

	do {
		// Wait until no other thread has access control on the common data
		WaitForSingleObject(hReadData, INFINITE);
		if (pvStartReadQueue == NULL) {
			switch (dwCurrentType)
			{
				case PRT_DOCUMENT_WIDTH:
					dwReadWaitFlag |= DOCUMENT_WIDTH;
					break;
				case PRT_MSR_MICR_READ_DATA:
					dwReadWaitFlag |= MSR_MICR_DATA;
					break;
				case PRT_MSR_WRITE_STATUS:
					dwReadWaitFlag |= MSR_WRITE_STATE;
					break;
				case PRT_PRINTING_UNIT_PARAMETERS:
					dwReadWaitFlag |= PRINTING_UNIT_PARAM;
					break;
				case PRT_MOD_ID:
					dwReadWaitFlag |= MOD_ID;
					break;
				
			}

			WakeUpReadThread();
			SetEvent(hReadData); // Allow other threads again to access common data
			// wait until the next information is read but not more than
			// WaitForSingleObject milli seconds
			dwRet = WaitForSingleObject(hWakeUp, MAX_WAIT_FOR_READ);  	
								// Now get access control on the common data again
			WaitForSingleObject(hReadData, INFINITE);
			switch (dwCurrentType)
			{
				case PRT_DOCUMENT_WIDTH:
					dwReadWaitFlag &= ~DOCUMENT_WIDTH;
					break;
				case PRT_MSR_MICR_READ_DATA:
					dwReadWaitFlag &= ~MSR_MICR_DATA;
					break;
				case PRT_MSR_WRITE_STATUS:
					dwReadWaitFlag &= ~MSR_WRITE_STATE;
					break;
				case PRT_PRINTING_UNIT_PARAMETERS:
					dwReadWaitFlag &= ~PRINTING_UNIT_PARAM;
					break;
				case PRT_MOD_ID:
					dwReadWaitFlag &= ~MOD_ID;
					break;
				
			}
		}
		
		if (pvStartReadQueue == NULL) {
			if (dwRet == WAIT_TIMEOUT) {
				// The read thread didn't get new information
								// Allow other threads again to access common data
				SetEvent(hReadData); 
				return(-1);						// We didn't the desired type 
			} else {
				SetEvent(hReadData); 
				continue;			
			}
		}

		dwType = pvStartReadQueue->dwType;

		if (dwType == dwCurrentType) {
			// This entry contains the desired type
			*pReadQueueEntry = pvStartReadQueue;
			// Remove this entry out of the queue, but don't delete the entry itself
			RemoveFirstReadQueueEntry();
		} else {
			// This entry has a different type. Remove it from the queue and
			// delete it
			DeleteFirstReadQueueEntry();
		}
		SetEvent(hReadData); // Allow other threads again to access common data
	} while (dwType != dwCurrentType);
	return(dwType);
}

/*++

Routine Description:
    Insert the current entry at the end of the read queue 

Arguments:
	pReadQueueEntry: Read queue entry, which has to be inserted.
		
Return Value:
	None
--*/

void InsertReadQueueEntry(ReadQueue *pReadQueueEntry) {
	ReadQueue	*pvOldEndReadQueue;
	DWORD		dwCounter = 0;
	
	// Save the information of the current last entry in the read queue
	pvOldEndReadQueue = pvEndReadQueue;
	if (pvOldEndReadQueue == NULL) {
		// The read queue was empty until now.
		// That is why the first and the last entry of the read queue
		// are now the same.
		pvEndReadQueue = (ReadQueue *)  pReadQueueEntry;
		pvStartReadQueue = (ReadQueue *)  pReadQueueEntry;
	} else {
		// The old last entry will now get the new entry as the next entry
		// in the read queue.
		pvOldEndReadQueue->pvNextReadQueue = (VOID *) pReadQueueEntry;
		// The new entry is now the last entry in the read queue.
		pvEndReadQueue = pReadQueueEntry;
	}

	return;
}

/*++

Routine Description:
    Calls PrtRead() to get the next information from the printer,
	if new information is available in the given timeout time.
	PrtRead() is called here again and again until all information available
	is returned. 
	This application waits the first time not more than one second (dwTimeout=1000 
	or PRT_IMMEDIATELY), if no information is currently available.
	All other call of PrtRead() don't wait, if no information is availabe 
	(PRT_IMMEDIATELY).

	(The PrtRead function gets status information from the printer, 
	 which the printer generates automatically as well as the answers 
	 to requests from the application.
	 The application is responsible to read until the answer string is received.
	 If no information is available after dwTimeout seconds, 
	 the function will return without any information.
	 A lot of strings returned from the printer are as follows:
         ESC [ P1; ... ;Pn <intermediate character><final character>
						   [<STX><text><ETX>]
	 with 
         P1 ... Pn numeric values
         STX: start of text
         ETX: end of text  
	 The numeric Parameter P1; ... ;Pn  after the control sequence introducer 
	 (CSI) ESC [ are returned to the application in the numeric array pdwParArray 
	 and the <text> between <STX> and <ETX> is returned in pbReadData, 
	 for those functions, which contain such <text>.
	 The string can also contain:
         ESC ID <module Ids/data Ids/electric journal function enabled>
		 (For instance:
				ESC ID$MOD$ 040125 0105 BOOTPROM.PRM:$MOD$031219 0208 4915_STD.MOD;$MOD$ ELJ
			  or
				ESC ID$MOD$ 040125 0105 BOOTPROM.PRM:$MOD$031219 0208 4915_STD.MOD;
			  It always ends with ';' or ";$MOD$ ELJ")
	    ESC MOD						(PRT_READY_TO_LOAD)
		ESC mod						(PRT_READY_TO_LOAD_XE)
		ESC FNT						(PRT_READY_TO_LOAD_FNT)
	 or
		ESC BEL L		(BEL=0x07)	(PRT_BEL_L).
	 The contents after ESC ID will be returned to the application in pbReadData.
	 If you want to get the information from the printer synchronized 
	 after every PrtWrite(), which asks for information from the printer, 
	 you can call PrtRead() after such a PrtWrite() in the same Thread. 
	 If you want to use PrtWrite() and PrtRead() asynchronous for some functions, 
	 you have to start an own thread, which we did in this application.
	 For functions like MICR read process or read a MSR track, 
	 where you want to get the information synchronized, 
	 although PrtRead() is called in a different thread, 
	 you have to wait after PrtWrite() for instance with WaitForSingleObject() 
	 until the PrtRead thread wakes you up for instance with SetEvent().)


Arguments:
	hPrinter:  Handle to the printer, which is returned from PrtOpen()
	dwTimeout: >0: Max. time in milli seconds to wait for the first PrtRead
			   PRT_IMMEDIATELY: Don't wait, if no information is available
			   PRT_INFINITE:	Wait until information is available	
Return Value:
	None
--*/

DWORD CallPrtRead(HANDLE hPrinter, DWORD	dwTimeout)
{
	ReadQueue	*pNewReadQueueEntry = NULL;
	DWORD		*pdwParArray = NULL;
	DWORD		dwPar[2];
	DWORD		dwArraySize = 2;
	DWORD		dwArrayLength;
	DWORD		dwReadDataLen = 0;
	DWORD		dwRet;
	DWORD		dwType;
	CHAR		*pcReadData = (CHAR *) NULL;
								// Start with the timeout time you got
	DWORD		dwCurrentTimeout = dwTimeout;
	
	// Call PrtRead() until the read queue of HPRDH.dll is empty
	for (;;) {
		// First only try to get the information, if not more than 2 variables
		// are in the variable list. In that case we will get the information
		// about the control points reached, the global and the special state.
		// If more variables are in the list and/or there is an additional text
		// we will get the information of the needed size in dwArraySize and/or
		// in dwReadDataLen.
		pdwParArray = &dwPar[0];
		dwPar[0] = 0;
		dwPar[1] = 0;
		// A wait time greater than 1 second doesn't make sense,
		// because the dll (HPRDH.dll) will retry to get the information
		// from the printer again and again during the time the application
		// is waiting in PrtRead(), if no information is returned.
		dwRet = PrtRead(hPrinter,	  // in: Printer handle (returned from PrtOpen()
									  // >0:			  Max. wait time in seconds 
									  // PRT_IMMEDIATELY: Return immeddiately
									  // PRT_INFINITE:	  Wait until information is 
						dwCurrentTimeout,			   // avaliable
						&dwType,	  // out: type of information (see HPRDH.h) 
						pdwParArray,  // out: array for the numerical variables 
						&dwArraySize, // out: Number of numerical variables returned
							// out: destination string for additional text contents.
							     // Here we took a NULL pointer, because we first  
						NULL,    // want to get the size needed for the string.
							// out: The number of bytes returned in the text string
						&dwReadDataLen); // or the number of bytes needed

		// If the communication to the printer is broken (printer is offline),
		// a PrtRelease() has to be done.
		// A PrtClaim has to be done, after the printer is online again.
		if (dwRet == PRT_COMMUNICATION_BROKEN) {
			PrtRelease(hPrinter);
		}

		// If no data is available, finish this function
		if ( (dwRet != ERROR_MORE_DATA) && (dwRet != PRT_NO_ERROR) ) {
			if ( (dwRet != ERROR_NO_MORE_ITEMS) && (dwRet != WAIT_TIMEOUT) ) {
				PrintMessage(FALSE, NULL, dwRet, "Read Printer");
			}
			break;
		}

		dwCurrentTimeout = PRT_IMMEDIATELY; // The next times don't wait

		switch (dwType) {

			// Control point is reached
			case PRT_CONTROL_POINT:
				WaitForSingleObject(hReadData, INFINITE);
				// Get the contents of the variable of check point reached
				lControlPointReached = (LONG) dwPar[0];
				if (dwReadWaitFlag & CONTROL_POINT) {
					if (lControlPoint2Wanted == lControlPointReached) {
											// Wake up the main thread, 
						SetEvent(hWakeUp);	// if it is waiting for new information
						dwReadWaitFlag &= ~CONTROL_POINT;
					}
				}
				// Wake up the main thread, if it is waiting for new information
				SetEvent(hWakeUp);
				SetEvent(hReadData); // Allow other threads again to access common data
				continue;

			// Control point 2 is reached
			case PRT_CONTROL_POINT_2:
				WaitForSingleObject(hReadData, INFINITE);
				// Get the contents of the variable of check point 2 reached
				lControlPoint2Reached = (LONG) dwPar[0];
				if (dwReadWaitFlag & CONTROL_POINT_2) {
					if (lControlPoint2Wanted == lControlPoint2Reached) {
											// Wake up the main thread, 
						SetEvent(hWakeUp);	// if it is waiting for new information
						dwReadWaitFlag &= ~CONTROL_POINT;
					}
				}
							// Allow other threads again to access common data
				SetEvent(hReadData); 
				continue;

			// The printer returned a global printer state.
			// You cannot ask the printer for a global printer state.
			// The printer returns this state, if the global state changes.
			case PRT_GLOBAL_STATUS:
				WaitForSingleObject(hReadData, INFINITE);
				if (dwArraySize > 0) {
					// Get the new global printer state
					lGlobalStatus = (LONG) dwPar[0];
				} else {
					// If no value is returned, the default value is online
					lGlobalStatus = GLOB_ONLINE;
				}
				if (dwReadWaitFlag & GLOBAL_STATE) {
					// Wake up the main thread, if it is waiting for new information
					SetEvent(hWakeUp);
					dwReadWaitFlag &= ~GLOBAL_STATE;
				}
							// Allow other threads again to access common data
				SetEvent(hReadData); 
				continue;

			// The printer returned a special printer message.
			// The special printer message is returned,
			// if the special state changes, 
			// whithout the printer.
			// But it is also possible to ask the printer for
			// the current printer state.
			case PRT_SPECIAL_MESSAGE:
				WaitForSingleObject(hReadData, INFINITE);
				// Get the first parameter of the special printer state.
				// 1: The paper sensor on the document printing point
				//    is covered with paper again.
				// 2: The paper sensor on the document printing point
				//    is no longer covered with paper
				// 20: USER 1 or USER 2 key is pressed
				lSpecialStatus1 = (LONG) dwPar[0];
				// Get the second parameter of the special printer state.
				// 1: USER 1 key is pressed
				// 2: USER 2 key is pressed
				// 4: The first parameter contains the current paper state
				lSpecialStatus2 = (LONG) dwPar[1];

				if (lSpecialStatus2 == SPC_PAPER_STATE) {
															// Is no paper entered ?
					if (lSpecialStatus1 == SPC_NOPAPER) {
																			// yes
						if (dwReadWaitFlag & NO_PAPER) {
							SetEvent(hWakeUp);
							dwReadWaitFlag &= ~NO_PAPER;
						}
					} else {
															  // Is paper inserted ?
						if (lSpecialStatus1 == SPC_PAPER) {
																			// yes
							if (dwReadWaitFlag & PAPER) {
								SetEvent(hWakeUp);
								dwReadWaitFlag &= ~PAPER;
							}
						}
					}
				} else {
					if (lSpecialStatus1 == SPC_USER_KEY) {
														 // Is USER key 1 pressed ?
						if (lSpecialStatus2 == SPC_USER_KEY1) {
																		   // yes
							if (dwReadWaitFlag & USER1_KEY) {
								SetEvent(hWakeUp);
								dwReadWaitFlag &= ~USER1_KEY;
							}
						} else {
														 // Is USER key 2 pressed ?
							if (lSpecialStatus2 == SPC_USER_KEY2) {
																		   // yes
								if (dwReadWaitFlag & USER2_KEY) {
									SetEvent(hWakeUp);
									dwReadWaitFlag &= ~USER2_KEY;
								}
							}
						}
					}
				}
							// Allow other threads again to access common data
				SetEvent(hReadData); 
				continue;
		}
	
		// Put all other returned types into a read queue entry and
		// add it to the read queue.

		// Allocate the needed space for the read queue entry
		pNewReadQueueEntry = (ReadQueue *) malloc(sizeof (ReadQueue));
		// Initialize the allocated entry with binary 0
		(void) memset (pNewReadQueueEntry, 0, sizeof (ReadQueue));
	
		if (dwArraySize > 0) {
			// Allocate the needed space for variables in the variable list.
			// The space has also to be allocated, if we already have got
			// all variables during the last PrtRead(),
			// because we have to put these variables in that case from the local
			// dwPar[2] into the global allocated space.
			dwArrayLength = dwArraySize * sizeof (DWORD);
			pdwParArray = (DWORD *) malloc(dwArrayLength);
			// Set the pointer on the array in the new read entry to the
			// just allocated space.
			pNewReadQueueEntry->pdwParArray = pdwParArray;
			// Initialize the allocated entry with binary 0
			(void) memset (pdwParArray, 0, dwArrayLength);
		}

		if (dwRet == ERROR_MORE_DATA) {
			// We have to call PrtRead() again to get all information now,
			// because now the needed space is or will be allocated.

			if (dwReadDataLen > 0) {
				// Additional text has to be returned.
				// Allocate the needed space for this information
				pcReadData = (CHAR *) malloc(dwReadDataLen + 1);
				// Set the corresponding pointer in the read queue entry on
				// the just allocated space.
				pNewReadQueueEntry->pcReadData = pcReadData;
				// Initialize the allocated entry with binary 0
				memset(pcReadData, (int) '\0', dwReadDataLen + 1);
			}
			// Now we can read with timeout time,
			// because information for us can be returned immediately
			dwRet = PrtRead(hPrinter, PRT_IMMEDIATELY, &dwType, 
							pdwParArray, &dwArraySize, 
							(BYTE *) pcReadData, &dwReadDataLen);
		} else {
			// We got all information during the last PrtRead().
			// We only have to insert it into the allocated array.

			if (dwArraySize > 0) {
				pdwParArray[0] = dwPar[0];
			}

			if (dwArraySize > 1) {
				pdwParArray[1] = dwPar[1];
			}
		}

		// Put the type into the new read queue entry
		pNewReadQueueEntry->dwType = dwType;
		// Put the size of the variable list into the new read queue entry
		pNewReadQueueEntry->dwNoOfElements = dwArraySize;
		// Put the length of the additional text into the new read queue entry
		pNewReadQueueEntry->dwReadDataLen = dwReadDataLen;
	
		if ( (dwRet != PRT_NO_ERROR) && (dwRet != WAIT_TIMEOUT) ) {
			// Print error message in a dialog box
			PrintMessage(FALSE, NULL, dwRet, "Read Printer");
			break;
		}

		// Wait until no other thread has access control on the common data
		WaitForSingleObject(hReadData, INFINITE);
		// Insert the new entry at the end of the read queue
		InsertReadQueueEntry(pNewReadQueueEntry);
		// Wake up the main thread, if it is waiting for new information

		switch (dwType) {

			// The width of the document is returned
			case PRT_DOCUMENT_WIDTH:
				if (dwReadWaitFlag & DOCUMENT_WIDTH) {
											// Wake up the main thread, 
					SetEvent(hWakeUp);	// if it is waiting for new information
					dwReadWaitFlag &= ~DOCUMENT_WIDTH;
				}
				break;

			// The result of an MSR or MICR read is returned
			case PRT_MSR_MICR_READ_DATA:
				if (dwReadWaitFlag & MSR_MICR_DATA) {
											// Wake up the main thread, 
					SetEvent(hWakeUp);	// if it is waiting for new information
					dwReadWaitFlag &= ~MSR_MICR_DATA;
				}
				break;
			
			// The printer's response on an MSR write
			case PRT_MSR_WRITE_STATUS:
				if (dwReadWaitFlag & MSR_WRITE_STATE) {
											// Wake up the main thread, 
					SetEvent(hWakeUp);	// if it is waiting for new information
					dwReadWaitFlag &= ~MSR_WRITE_STATE;
				}
				break;
			
			// The printing unit parameters are returned
			case PRT_PRINTING_UNIT_PARAMETERS:
				if (dwReadWaitFlag & PRINTING_UNIT_PARAM) {
											// Wake up the main thread, 
					SetEvent(hWakeUp);	// if it is waiting for new information
					dwReadWaitFlag &= ~PRINTING_UNIT_PARAM;
				}
				break;

			// The desired data IDs of the character generators are returned.
			case PRT_MOD_ID:
				if (dwReadWaitFlag & MOD_ID) {
											// Wake up the main thread, 
					SetEvent(hWakeUp);	// if it is waiting for new information
					dwReadWaitFlag &= ~MOD_ID;
				}
				break;
		}
	
		SetEvent(hReadData);	// Allow other threads again to access common data
	}
	
	return(dwRet);
}


/*++

Routine Description:
    This is the read thread, which is calling PrtRead() via CallPrtRead().
	
Arguments:
	hPrinter: Handle to the printer, which is returned from PrtOpen()
		
Return Value:
	None
--*/

void AsyncReadThread(HANDLE hPrinter) {
	DWORD	dwRet = WAIT_TIMEOUT;
	DWORD	dwTimeout = PRT_IMMEDIATELY; // Normally don't wait during PrtRead()
	
	for (;;) {
		// Wait until no other thread has access control on the common data
		WaitForSingleObject(hReadData, INFINITE);

		if (    (dwRet != PRT_NO_ERROR)
			 && (dwRet != ERROR_NO_MORE_ITEMS) 
 		     && (dwRet != WAIT_TIMEOUT) ) {
			dwRetLastRead = dwRet;
			StopThread = FALSE;
			SetEvent(hReadData); // Allow other threads again to access common data
			PrtClose(&hPrinter);
			exit(dwRet);
		}

		if (dwReadWaitFlag == 0) {
			SetEvent(hReadData);
			dwRet = WaitForSingleObject(hWakeupThread, WAIT_TIME_READ_THREAD);
			WaitForSingleObject(hReadData, INFINITE);
		}
		if (dwReadWaitFlag) {
			// We have been waked up from the main thread to call PrtRead()
			
			// Now wait during PrtRead(), if no data is available, because
			dwTimeout = WAIT_TIME_READ; // we expect information from the printer
			// Initialize, that the we will wait during the next
			// WaitForSingleObject(hWakeupThread);
			ResetEvent(hWakeupThread);
		} else {
			dwTimeout = PRT_IMMEDIATELY; // Normally don't wait during PrtRead()
		}

		if (StopThread) {
			// The main thread wants, that the read threads stops.

			// Initialize variable again for later use
			StopThread = FALSE;
			SetEvent(hReadData); // Allow other threads again to access common data
			ExitThread(0);
		}
		SetEvent(hReadData);	// Allow other threads again to access common data
								// Try to get new information from printer
		dwRet = CallPrtRead(hPrinter, // in: Printer handle (returned from PrtOpen())		
							dwTimeout); // in: wait time in seconds for PrtRead()

	}
} 


/*++

Routine Description:
    Creates the ReadThread

Arguments:
    
Return Value:
    TRUE on succesfully creating the thread, else FALSE

Hint:
	Use AfxBeginThread() in MFC programs;

	pThreadObject = AfxBeginThread((AFX_THREADPROC) ReadThread, (LPVOID) this, 
								   THREAD_PRIORITY_NORMAL, 0, 0, NULL);

	UINT ReadThread(LPVOID pDialogObject) {
		CHprApplicationDlg *pHprApplObj = (CHprApplicationDlg *) pDialogObject;

		pHprApplObj->AsyncReadThread();
		return 0;
	}

--*/

void CreateReadThread() {

    HANDLE  ThreadHandle;
    DWORD   ThreadId;
 
	// Create the thread, which has to call PrtRead() asynchron
	ThreadHandle = CreateThread(NULL, 0,
											// in: start routine of read thread
                                (LPTHREAD_START_ROUTINE) AsyncReadThread,
								// in: parameter for read thread (Printer handle)
                                hPrinter, 
								0, 
								&ThreadId);	// out: The thread id of the read thread
    if (ThreadHandle) {
		// The thread is created, but we don't need the corresponding handle
		// any more.
		CloseHandle(ThreadHandle);
        return;
    } else {
		// Can't create the read thread
		PrintMessage(FALSE, NULL, GetLastError(), "Create read thread");
		exit (1);	// Without a read thread, we can't work correct
	}
}

/*++

Routine Description:
    Is printer online.
	If not start dialog thread to print the corresponding error message
	and wait until the printer is online again or the user enters cancel.

Arguments:
	CheckPaper: TRUE: Check also, if paper is available
		
Return Value:
	None
--*/

void PrinterOnlineAndPaper(BOOL CheckPaper) {
	char *pcText;
	BOOL OnlineAndPaper = TRUE;
	
	WaitForSingleObject(hReadData, INFINITE);

	// Printer is not online (stopped or error)
	if ( (lGlobalStatus != GLOB_ONLINE) && (lGlobalStatus != GLOB_UNDEF) ) {
		if (lGlobalStatus == GLOB_STOP) {
			pcText = "Printer is stopped";
		} else {
			pcText = "Printer error!";
		}
		OnlineAndPaper = FALSE;
	} else {
		if (	(CheckPaper == TRUE) 
			 && (lSpecialStatus1 == SPC_NOPAPER)
			 && (lSpecialStatus2 == SPC_PAPER_STATE) ) {
			pcText = "No paper";
			OnlineAndPaper = FALSE;
		}
	}
	if (OnlineAndPaper == FALSE) {
		SetEvent(hReadData);

		if (CreateDialogThread(pcText) == TRUE) {
			while (TRUE) {
				if (DlgResult == DLG_CANCEL) {
					// The user pressed the cancel button
					EndPrinting();
					exit(2);
				}
				// Wait until no other thread has access control on the common data
				WaitForSingleObject(hReadData, INFINITE);
				if (CheckPaper == TRUE) {
					if (    (lGlobalStatus == GLOB_ONLINE)
						 && (lSpecialStatus1 == SPC_PAPER)
						 && (lSpecialStatus2 == SPC_PAPER_STATE) ) {
						break;
					}
				} else {
					if (lGlobalStatus == GLOB_ONLINE) {
						break;
					}
				}
						// wake up the thread to do the next read, because we wait
				SetEvent(hWakeupThread); // printer online
				ResetEvent(hWakeUp);		
				dwReadWaitFlag |= GLOBAL_STATE;
				SetEvent(hReadData);
					// Wait until the read thread informs us, that new information
				WaitForSingleObject(hWakeUp, MAX_WAIT_FOR_READ);	// arrived
				// Wait until no other thread has access control on the common data
				WaitForSingleObject(hReadData, INFINITE);
				dwReadWaitFlag &= ~GLOBAL_STATE;
				SetEvent(hReadData); 
			}
			if ((!DlgResult) && (DlgHwnd != (HWND)0)) {	   // dialog still active ?
				// End the dialog 
				(void)EndDialog (DlgHwnd,(WPARAM) DLG_END);
				DlgHwnd = NULL;
			}
		}
	}
	SetEvent(hReadData);
	return;
}
	

/*++

Routine Description:
    Call PrtWrite() to write the given string to the printer

Arguments:
	pcWriteString: The string, which has to be written
	pdwNoOfBytesWritten: in:  the number of bytes to be written
						 out: the number of bytes written
	dwRetAllowed:  Error code, which doesn't end printing
	dwRetAllowed2: Error code, which doesn't end printing
	CheckPaper: TRUE: Check also, if paper is available
		
Return Value:
	The return code of PrtWrite()
--*/

int WriteToPrinter(char *pcWriteString, LPDWORD pdwNoOfBytesWritten, 
				   DWORD dwRetAllowed, DWORD dwRetAllowed2, BOOL CheckPaper) {
	DWORD	dwRet;
	int		LenWriteString;
	char	Message[ERROR_STRING_SIZE + 1];
	BOOL	MessageThreadStarted = FALSE;
	char	*pcWriteString2 = pcWriteString;
	DWORD	dwNoOfBytesWritten = 0;
	
	LenWriteString = *pdwNoOfBytesWritten;	// The number of bytes to be written
	*pdwNoOfBytesWritten = 0;
	PrinterOnlineAndPaper(CheckPaper);
	while (TRUE) {
								/* User entered cancel to stop Dialog thread ? */
		if ( (MessageThreadStarted == TRUE) && (DlgResult == DLG_CANCEL) ) {		    /* yes */
			EndPrinting();		// Stop read thread, PrtRelease() and PrtClose()
			exit(3);
		}

										  // write the string via COM1, COM2 or USB 
		dwRet = PrtWrite(hPrinter,			  // in: Handle returned from PrtOpen()
						(LPBYTE) pcWriteString2,	// in: Contents to be written
						LenWriteString,				// in: Number of bytes to write
						&dwNoOfBytesWritten);		// out: Number of bytes written

		// If the communication to the printer is broken (printer is offline),
		// a PrtRelease() has to be done.
		// A PrtClaim has to be done, after the printer is online again.
		if (dwRet == PRT_COMMUNICATION_BROKEN) {
			PrtRelease(hPrinter);
		}

		if (    (dwRet != PRT_NO_ERROR) 
			 && (dwRet != dwRetAllowed) 
			 && (dwRet != dwRetAllowed2) ) {
			if (MessageThreadStarted == FALSE) {
				PrintMessage(TRUE, Message, dwRet, "Write Printer");
				MessageThreadStarted = TRUE;
			}
			continue;
		} else {

			// Nach einem erfolgreichen Schreiben das Dialog-Fenster wieder
			// schlieen, sofern ein Dialog in einem anderen Thread angezeigt wird.

			if ((!DlgResult) && (DlgHwnd != (HWND)0)) {	   // dialog still active ?
				// End the dialog 
				(void)EndDialog (DlgHwnd,(WPARAM) DLG_END);
				DlgHwnd = NULL;
			}

			/* Hier wird berfprft werden, ob alle Zeichen geschrieben wurden.
			 * Die noch nicht geschriebenen Zeichen werden erneut geschrieben. */
			 
			*pdwNoOfBytesWritten += dwNoOfBytesWritten;
			LenWriteString -= dwNoOfBytesWritten;
			if (LenWriteString > 0) {
				pcWriteString2 += dwNoOfBytesWritten;
				dwNoOfBytesWritten = 0;
				continue;
			}
			return(dwRet);
		}
	}
	return(dwRet);
}

/*++

Routine Description:
    Wait for the desired control point.
	This control point is also reached, 
	if the value is meanwhile greater than the desired value

Arguments:
	dwControlPoint: The control point, to wait for
		
Return Value:
	The control point reached meanwhile.
--*/

DWORD WaitForControlPoint(LONG lControlPoint, bool bControllPoint2) {
	LONG	lCurrControlPoint = -1;	// Save the control point reached last time
	LONG	lCurrControlPointNew = -1;
	DWORD	dwRet;

	do {
									  // Did we reach the desired control point 
		if (lCurrControlPoint < lControlPoint) {				// or even more ?
																		  // no 
				// Wait until no other thread has access control on the common data
			WaitForSingleObject(hReadData, INFINITE);
										// Save the control point reached until now
			if (bControllPoint2 == true) {
				lCurrControlPoint = lControlPoint2Reached;
			} else {
				lCurrControlPoint = lControlPointReached;
			}
							// Did we meanwhile reach the desired control point 
			if (lCurrControlPoint < lControlPoint) {			// or even more ?
																		  // no
				if (bControllPoint2 == true) {
					dwReadWaitFlag |= CONTROL_POINT_2;
					lControlPoint2Wanted = lControlPoint;
				} else {
					dwReadWaitFlag |= CONTROL_POINT;
					lControlPointWanted = lControlPoint;
				}

				WakeUpReadThread();

				SetEvent(hReadData);
				// Wait not more than MAX_WAIT_FOR_READ milli seconds to be woken
				// up from the read thread, if new information arrived
				dwRet = WaitForSingleObject(hWakeUp, MAX_WAIT_FOR_READ);
						// Init value, that this waits waits during next call of
				WaitForSingleObject(hReadData, INFINITE);
				if (bControllPoint2 == true) {
					dwReadWaitFlag &= ~CONTROL_POINT_2;
					lCurrControlPointNew = lControlPoint2Reached;
				} else {
					dwReadWaitFlag &= ~CONTROL_POINT;
					lCurrControlPointNew = lControlPointReached;
				}
				if (    (dwRet == WAIT_TIMEOUT) 
					 && (lCurrControlPointNew == lCurrControlPoint) ) {
																			 // no
								// Allow other threads again to access common data
					SetEvent(hReadData);
					return(lCurrControlPoint);			  // return without success
				}
										// Save the control point reached until now
				lCurrControlPoint = lCurrControlPointNew;
			}
			SetEvent(hReadData); // Allow other threads again to access common data
		}
	} while (lCurrControlPoint < lControlPoint);
						// return the desired control point or even a newer one	
	return(lCurrControlPoint);	
}

/*++

Routine Description:
    Wait until paper is inserted or Cancel is entered from the user.

Arguments:
	None
Return Value:
	true:  paper is inserted
	false: no paper
--*/

bool WaitForPaper() {
	bool	GotPaper = false;
	LONG	lCurrControlPoint = -1;
	
	// Wait for paper

	// Start a thread, which asks the user to enter new paper.
	// If paper is entered, the dialog is aborted.
	// If no paper is entered and the user presses the cancel button,
	// the program ends.
	if (CreateDialogThread(NULL) == TRUE) {
		while (TRUE) {
			if (DlgResult == DLG_CANCEL) {
				// The user didn't enter paper. He pressed the cancel button
				break;
			}
				// Wait until no other thread has access control on the common data
			WaitForSingleObject(hReadData, INFINITE);
			if (lGlobalStatus == GLOB_ERROR) {
				// The printer is not online and is not stopped
								// Allow other threads again to access common data
				SetEvent(hReadData);
				break;							  // Don't wait for paper any more
			} else {
				
				if (	(lSpecialStatus1 == SPC_PAPER)
					 && (lSpecialStatus2 == SPC_PAPER_STATE) ) {

					// Paper exists

					if (lGlobalStatus != GLOB_STOP) {
						// The printer is also online
						GotPaper = true;	// We now have paper to print
					}
								// Allow other threads again to access common data
					SetEvent(hReadData);
					break;
				}
			}
			dwReadWaitFlag |= GLOBAL_STATE;
			dwReadWaitFlag |= PAPER;
			WakeUpReadThread();
			SetEvent(hReadData); // Allow other threads again to access common data
					// Wait until the read thread informs us, that new information
			WaitForSingleObject(hWakeUp, MAX_WAIT_FOR_READ);	// arrived
				// Wait until no other thread has access control on the common data
			WaitForSingleObject(hReadData, INFINITE);
			dwReadWaitFlag &= ~GLOBAL_STATE;
			dwReadWaitFlag &= ~PAPER;
			SetEvent(hReadData); // Allow other threads again to access common data
		}
	}

	if ((!DlgResult) && (DlgHwnd != (HWND)0)) {			// dialog still active ?
		// If paper is inserted, we have to end the dialog for paper
		(void)EndDialog (DlgHwnd,(WPARAM) DLG_END);
		DlgHwnd = NULL;
	}
	return(GotPaper);
}

/*++

Routine Description:
    Initialize the printer etc.
Arguments:
	none
Return Value:
	PRT_NO_ERROR: OK
	<> 0:	Error returned from PrtWrite()
--*/

DWORD InitializePrinter()
{
	DWORD		dwRet;
	DWORD		dwNoOfBytesWritten = 0;

	// Initialize the printer
	dwNoOfBytesWritten = strlen(INIT_PRINTER);
	dwRet = WriteToPrinter(INIT_PRINTER, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Select graphical character set
	dwNoOfBytesWritten = strlen(GRAPH_CHAR_SET);
	dwRet = WriteToPrinter(GRAPH_CHAR_SET, (LPDWORD) &dwNoOfBytesWritten, 
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Switch on extended character set
	dwNoOfBytesWritten = strlen(EXT_CHAR_SET);
	dwRet = WriteToPrinter(EXT_CHAR_SET, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Select the character set of the USA
	dwNoOfBytesWritten = sizeof(US_CHAR_SET);
	dwRet = WriteToPrinter(US_CHAR_SET, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Setting the left margin to 1
	dwNoOfBytesWritten = strlen(LEFT_MARGIN10);
	dwRet = WriteToPrinter(LEFT_MARGIN10, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Switch to LQ print quality
	dwNoOfBytesWritten = strlen(LQ_PRINT_QUALITY);
	dwRet = WriteToPrinter(LQ_PRINT_QUALITY, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Select 1/6 inch line spacing
	dwNoOfBytesWritten = strlen(INCH_LINE_SPACE);
	dwRet = WriteToPrinter(INCH_LINE_SPACE, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Determing the document length in lines to 60
	dwNoOfBytesWritten = strlen(DOC_LEN_IN_LINES_60);
	dwRet = WriteToPrinter(DOC_LEN_IN_LINES_60, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Absolute vertical pos. from the top (100)
	dwNoOfBytesWritten = strlen(VERT_POS_100);
	dwRet = WriteToPrinter(VERT_POS_100, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	// Absolute horizontal positioning 
	dwNoOfBytesWritten = sizeof(HORIZ_POS);
	dwRet = WriteToPrinter(HORIZ_POS, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	return(dwRet);
}
/*++

Routine Description:
    Insert the current entry at the end of the read queue 

Arguments:
	pcPrinterName: The name of the printer
		
Possible return values from the printer functions (PrtOpen(), PrtClose() etc):
	PRT_NO_ERROR: The printing was successful
	ERROR_INVALID_PRINTER_NAME (1801): The printer name was not a valid printer 
									   (PrtOpen())
	ERROR_NOT_READY (21): The printer is not ready yet or not ready any more.
						( - The print spooler closed the corresponding port,
							because the configuration of the printer was changed
							between PrtOpen() and PrtClaim()
						  - The desired control points couldn't be reached,
							because the printer is not online any more.)
						The user has to eliminate the problem. Then printing
						can start again.
	ERROR_BAD_LENGTH (24): - The printer name contains more than 256 character 
							 (PrtOpen())
	ERROR_OUT_OF_PAPER (28): - The printer is out of paper (USB). All other return
							   ERROR_TIMEOUT in that case.
	ERROR_MORE_DATA	(234): - The PrtRead cannot return the data, because the array 
							 or the string is not big enough. 
							 (The needed size for array and/or string is returned.)
	ERROR_NOT_ENOUGH_MEMORY (8): HPRDH.dll could not allocate the desired space
	WAIT_TIMEOUT (258): The called function was not successful until the timeout
					    was reached. (PrtRead(), PrtClaim() and PrtWrite())
	ERROR_TIMEOUT (1460): PrtWrite() returned because the timeout period expired.
						  (If the printer is stopped an ERROR_TIMEOUT can occur) 	
	ERROR_NO_MORE_ITEMS (259): PrtRead() is called without a timeout time 
							   (PRT_IMMEDIATELY) and no information is available
	RPC_S_SERVER_UNAVAILABLE (1722): The RPC server is unavailable
	RPC_S_CALL_FAILED (1726): The remote procedure call failed 
							  (The print spooler may be stopped.)
	RPC_S_CALL_FAILED_DNE (1727): The remote procedure call failed 
								  and did not execute. 
 								  (The print spooler may be stopped.)	

	The next error codes are the result of programming errors in the application:

	ERROR_INVALID_HANDLE (6):	  The printer handle is not correct any more
	ERROR_INVALID_PARAMETER (87): The functions are called with an invalid parameter
	ERROR_BUSY (170):	- Printer is already connected (PrtOpen())
						- PrtClaim() is called twice without a PrtRelease() between
	PRT_OPEN_NOT_DONE ((0x20000001):   PrtClaim, PrtRead, PrtWrite, PrtRelease or
								   PrtClose is called without calling PrtOpen first
	PRT_CLAIM_NOT_DONE (0x20000002):  PrtRead, PrtWrite or PrtRelease is called 
									  without calling PrtClaim first
	PRT_COMMUNICATION_BROKEN (0x20000003): 
							 The communication with the printer is broken
							 (A PrtRelease() has to be done at once and a 
							  a PrtClaim() after the printer is online again.
							  (PrtClaim(), PrtRead() and PrtWrite()). 
							   With patch WindowsXP-KB834462-x86-DEU.EXE installed
							   the PrtRelease() can also be done after the printer
							   is online again.)

    An error different from ERROR_BAD_LENGTH, WAIT_TIMEOUT or ERROR_NO_MORE_ITEMS
	has to stop reading from printer (PrtRead).

	All other error codes:	 error codes returned from functions called from
							 HPRDH.dll or the language monitor (wnhprlm.dll)

					 
--*/

int DoPrinting(char *pcPrinterName) {
	DWORD		dwRet;
	DWORD		dwCounter = 0;
	DWORD		*pdwParArray; 
	DWORD		dwNoOfBytesWritten = 0;
	bool		GotPaper = false;
	LONG		lCurrControlPoint;
	char		WriteString[MAX_WRITE_STRING];
	ReadQueue	*pReadQueueEntry = NULL;
	int			Line;
	int			Page;
	
	for (dwCounter = 0; dwCounter < NUMBER_OF_RETRIES; dwCounter++) {

		// Open the printer
							  // out: pointer to the handle is returned.
							  //      This handle is needed in all other	
		dwRet = PrtOpen(&hPrinter, // following calls (PrtClaim(), PrtWrite() etc)
							// in: Name of the printer (Start->Settings->Printer)
							//     (The bidirectional printing must be supported
							//      for that printer (Start->Settings->Printer->
						pcPrinterName);			   // Features->Connections))
		
		if (dwRet == ERROR_NOT_READY) {
			// The language monitor wasn't called with open port from the
			// spooler during connection from HPRDH.dll to LM.
			// We have to try again, because open port should be called
			// from the print spooler soon.
			// This error should normally not occur, because HPRDH.dll
			// retries also in that case.

			Sleep(WAIT_TIME_RETRY); // Wait half a second and then try again
			continue;
		}
		break;
	}

	if (dwRet != PRT_NO_ERROR) {
		PrintMessage(FALSE, NULL, dwRet, "Open Printer");
		return (dwRet);

	}

	for (dwCounter = 1; dwCounter <= NUMBER_OF_RETRIES; dwCounter++) {

		// Insert our job in the print queue and wait until our job is allowed
		// to print. Don't wait longer than CLAIM_TIMEOUT milliseconds for it.
		dwRet = PrtClaim(hPrinter,  // Handle to the printer (returned from PrtOpen)
							// The language monitor waits until the print spooler
							// informs him, that printing for this job can be
							// started. If other jobs are in the print queue before
							// our own job, it can be, that one of these jobs is
							// still active, when the timeout time is reached
						 CLAIM_TIMEOUT);
		
		if (dwRet == PRT_NO_ERROR) {
			// We can print now. First start the read thread, which is
			// responsible for the information returned from the printer.
			CreateReadThread();
			break;
		} else {
			if ( (dwRet == WAIT_TIMEOUT) && (dwCounter < NUMBER_OF_RETRIES) ) {
				// Try again, because our job was still in the queue
				// after CLAIM_TIMEOUT milliseconds. The next time our job
				// will be inserted again in the print queue. It may be
				// that meanwhile new jobs are inserted in that queue
				// after our old job and before our new job.
				// The old job will be ignored from LM,
				// even if this job is ready to print during the next PrtClaim()
				continue;
			}

			// It is possible, that the print spooler closed our port.
			// In that case ERROR_NOT_READY is returned.
			// We have to close the printer then and try again.
			// All other case lead to exit of this application.

			PrtClose(&hPrinter);						// Close the printer
			PrintMessage(FALSE, NULL, dwRet, "Claim Printer");
			return(dwRet);	// Return to the main function to retry or exit
		}
	}

	// The claim was successful: First clear the read queue,
	// because we are not interested in old printer information
	ClearReadQueue();

	// Eject the paper to be sure that printing is not started
	// on paper, which contains print lines from the last printing.
	dwNoOfBytesWritten = strlen(EJECT);
	dwRet = WriteToPrinter(EJECT, (LPDWORD) &dwNoOfBytesWritten, ERROR_TIMEOUT,
						   ERROR_OUT_OF_PAPER, FALSE);
	if ((dwRet == ERROR_TIMEOUT) || (dwRet == ERROR_OUT_OF_PAPER)) {
	
		// The printer wants to write something from the last job first

		GotPaper = WaitForPaper();			// Wait for paper
		if (GotPaper == true) {
			// Eject the paper to be sure that printing is not started
			// on paper, which contains print lines from the last printing.
			dwNoOfBytesWritten = strlen(EJECT);
			dwRet = WriteToPrinter(EJECT, (LPDWORD) &dwNoOfBytesWritten, 
								   ERROR_TIMEOUT, ERROR_OUT_OF_PAPER, FALSE);
			if (    (dwRet != PRT_NO_ERROR) 
				 && (dwRet != ERROR_TIMEOUT) 
				 && (dwRet != ERROR_OUT_OF_PAPER) ) {
				return(dwRet);
			}
		}
	} else {
	
		if (dwRet != PRT_NO_ERROR) {
			return(dwRet);
		}
	}

	// Set control point with value 1, because we want to be informed, 
	// if eject of paper is executed.
	sprintf(WriteString, SETCTLP, (int) 1);
	dwNoOfBytesWritten = strlen(WriteString);
	dwRet = WriteToPrinter(WriteString, (LPDWORD) &dwNoOfBytesWritten, 
						   ERROR_TIMEOUT, ERROR_OUT_OF_PAPER, FALSE);
	if (    (dwRet != PRT_NO_ERROR) 
		 && (dwRet != ERROR_TIMEOUT) 
		 && (dwRet != ERROR_OUT_OF_PAPER) ) {
		return(dwRet);
	}

	if (dwRet == PRT_NO_ERROR) {
		// Wait until the desired control point is reached
		lCurrControlPoint = WaitForControlPoint((LONG) 1, false);
	
		// A control point is reached, if the value returned is >= the desired value,
		// if we don't wait for every control point.
		if (lCurrControlPoint < 1) {
			EndPrinting();			// Stop read thread, PrtRelease() and PrtClose()
			return(ERROR_NOT_READY); // Perhaps try again after printer is online again
		}
		GotPaper = false;
	} else {
		// The printer may be stopped.
		// In that case we get ERROR_OUT_OF_PAPER from USB printer

		GotPaper = WaitForPaper();			
		
		if (GotPaper == false) {
			EndPrinting();		// Stop read thread, PrtRelease() and PrtClose()
			return(PRT_NO_ERROR);
		}
	}

	
	// First initialize the printer etc., before we really write contents
	dwRet = InitializePrinter();

	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}
	
	for (Page = 1; Page <= NUMBER_OF_PAGES; Page++) {
	
		if (GotPaper == false) {
			GotPaper = WaitForPaper();			
		
			if (GotPaper == false) {
				EndPrinting();		// Stop read thread, PrtRelease() and PrtClose()
				return(PRT_NO_ERROR);
			}
		}
		GotPaper = false;
	
		// Now some lines are written and a control point is set
		// after each line
		for (Line = 1; Line <= NUMBER_OF_LINES; Line++) {
			sprintf(WriteString, "\011This is line %d.\r\n", Line);
			dwNoOfBytesWritten = strlen(WriteString);
			dwRet = WriteToPrinter(WriteString, (LPDWORD) &dwNoOfBytesWritten,
								   PRT_NO_ERROR, PRT_NO_ERROR, TRUE);
			if (dwRet != PRT_NO_ERROR) {
				return(dwRet);
			}
		}

		WaitForSingleObject(hReadData, INFINITE);
		dwReadWaitFlag |= GLOBAL_STATE;
		dwReadWaitFlag |= NO_PAPER;

		WakeUpReadThread();
		SetEvent(hReadData);
		// stop execution, if printer is not online
		PrinterOnlineAndPaper(FALSE);	

		// Form Feed
		dwNoOfBytesWritten = strlen(FF);
		dwRet = WriteToPrinter(FF, (LPDWORD) &dwNoOfBytesWritten,
							   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
		if (dwRet != PRT_NO_ERROR) {
			return(dwRet);
		}
		
		// Wait until paper is ejected,
		// because we don't want to ask the user to enter paper
		// as long as the printer is writing the last page
		dwRet = WaitForSingleObject(hWakeUp, MAX_WAIT_FOR_EJECT);
		WaitForSingleObject(hReadData, INFINITE);
		dwReadWaitFlag &= ~NO_PAPER;
		dwReadWaitFlag &= ~GLOBAL_STATE;
		SetEvent(hReadData);
		PrinterOnlineAndPaper(FALSE);	// stop execution, if printer is not online
		if (dwRet == WAIT_TIMEOUT) {
			PrintMessage(FALSE, NULL, (DWORD) 0, "Paper not ejected");
		}
	}

	// Get the printing unit parameters
	
	dwNoOfBytesWritten = strlen(PRINTING_UNIT_PARAMETERS);
	dwRet = WriteToPrinter(PRINTING_UNIT_PARAMETERS, (LPDWORD) &dwNoOfBytesWritten,
						   PRT_NO_ERROR, PRT_NO_ERROR, FALSE);
	if (dwRet != PRT_NO_ERROR) {
		return(dwRet);
	}

	
	// Wait for the printing unit parameters
	GetInformationOfDesiredType(PRT_PRINTING_UNIT_PARAMETERS, &pReadQueueEntry);

					// Did we get read information belonging to the MICR read ?
	if (pReadQueueEntry != NULL) {
																		   // yes
		pdwParArray = &(pReadQueueEntry->pdwParArray[0]);

		// Print the printing unit parameters

		PrintMessage(FALSE, NULL, (DWORD) 0, 
		   		     "Printing unit parameters:\n\n%s station, %s, \nwidth in 1/10 inch: %ld, \nnumber of feed channels: %ld, \nnumber of eject channels: %ld, \nOCR reader not available, \nMSR %s, \nprinter type: %ld\n", 
					 (pdwParArray[0] == 0) ? "Document" : "Journal", 
					 (pdwParArray[1] == 3) ? "journal" : "document",
					 pdwParArray[2], 
					 pdwParArray[3],
					 pdwParArray[4],
					 (pdwParArray[6] == 0) ? "not available" : "available",
					 pdwParArray[7]);
		
		// Now delete the read queue entry
		// The entry is not in the read queue any more 
		// and has to be deleted, if the entry is not needed any more
		DeleteReadQueueEntry(pReadQueueEntry);
	} else {
		PrintMessage(FALSE, NULL, (DWORD) 0, 
					 "Printing unit parameters: No data from printer");
	}

	EndPrinting();				// Stop read thread, PrtRelease() and PrtClose()
	return(PRT_NO_ERROR);
}
