// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright 2004 Wincor Nixdorf.  All Rights Reserved.

// DemoApplication.cpp : Call: DemoApplication [<Printer name>]
// This program links HPRDH.lib and includes HPRDH.h to be able
// to use HPRDH.dll for the HighPrint 4915.
// In main() DoPrinting() in printer.cpp is called to do the printing.
//

#include "stdafx.h"
#include <windows.h>
#include <winspool.h>
#include <winioctl.h>
#include <winuser.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "HPRDH.h"			// Include file for the printing to the HighPrint 4915
#include "resource.h"

#define	ERROR_STRING_SIZE		1024
#define DEFAULT_PRINTER_NAME	"Wincor Nixdorf HighPrint 4915"
#define MAX_PRINTER_NAME_SIZE	256


					// In DoPrinting() etc. in printer.cpp all the printing is done
extern int DoPrinting(char *pcPrinterName);	

HANDLE  hModule;			// Handle to the module DemoApplication.exe
HANDLE	hDonePrintMessage;	// Handle to sequence the message windows of the threads
	// Handle to wake up the read thread, if information from the printer is wanted
HANDLE	hWakeupThread;	// or if the thread has to end
		// Handle to sequence access to the common data of main and read thread
HANDLE	hReadData;
		// Handle to wake up the main thread from the read thread, if information 
HANDLE	hWakeUp;	// from the printer is available
HWND	DlgHwnd;	// Handle to the dialog box, which asks to insert paper
		// Contains the information, if the user has pressed the cancel button 
DWORD	DlgResult;	// of the paper dialog
					// Printer name (default "Wincor Nixdorf HighPrint 4915")
char	PrinterName[MAX_PRINTER_NAME_SIZE + 1];  
		// The main thread sets StopThread to TRUE, if the read thread has to end

BOOL	StopThread; 

void usage(void)
{
    printf ("\nUsage DemoApplication [<Printer name>]\n");
    return;
}

// Callback routine for the paper and message dialogs

BOOL CALLBACK ToolDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lpText)
{
	switch(Message)
	{
	  	case WM_INITDIALOG:
			// Save handle to the dialog box (needed for the paper dialog)
			DlgHwnd = hwnd;
			// Write the printer name into the window text
			SetWindowText (hwnd,(LPSTR) PrinterName);
			if (lpText != NULL) {
				// Write the message, because "enter paper" is default
				SetDlgItemText(hwnd, IDC_MessageText, (LPSTR) lpText);
			}
			break;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDCANCEL:
					EndDialog(hwnd, (WPARAM)DLG_CANCEL);
					DlgHwnd = NULL;
					break;
			}
			break;
        case WM_CLOSE:
			 EndDialog(hwnd,(WPARAM)DLG_END);
			 DlgHwnd = NULL;
		     break;
		default:
			return FALSE;
	}
	return TRUE;
}

// The dialog box to enter paper is called in a different thread,
// because the main thread will continue at once, if the paper is inserted.
VOID DialogThread(char *pcText) {

	INT_PTR retDialog;
	
	retDialog = DialogBoxParam ((HINSTANCE) hModule, 
						   MAKEINTRESOURCE(IDD_TOOLBAR), NULL, ToolDlgProc,
						   (LPARAM) pcText);
	if (retDialog > 0) {
		// The dialog box was shown successfully
		DlgResult = retDialog;
		if (DlgResult == DLG_CANCEL) {
			// The user pressed the cancel button instead of entering paper.
			// Wake up the main thread, if it is waiting for paper.
			SetEvent(hWakeUp);
		}
	} else {
		DlgResult = DLG_ERROR;
	}
	DlgHwnd = NULL;
	return;
}


// Callback routine of the printer name dialog
// lpPrinterName: The default printer name
BOOL CALLBACK GetPrinterNameProc(HWND hwnd, UINT Message, WPARAM wParam, 
								 LPARAM lpPrinterName)
{
	switch(Message)
	{
	  	case WM_INITDIALOG:
			// Set dialog item text to the default printer name
			SetDlgItemText(hwnd, IDC_PrinterName, (LPSTR) lpPrinterName);
			break;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDCANCEL:
					EndDialog(hwnd, (WPARAM)DLG_CANCEL);
					exit(0);	// Exit execution of this process

				case IDOK:
					// Get the new printer name
					GetDlgItemText(hwnd, IDC_PrinterName, PrinterName, 
								   MAX_PRINTER_NAME_SIZE);
					EndDialog(hwnd, (WPARAM)DLG_CANCEL);
				break;
			}
			break;
        case WM_CLOSE:
			EndDialog(hwnd,(WPARAM)DLG_END);
			exit(0);
		    break;
		default:
			return FALSE;
	}
	return TRUE;
}

// Call the dialog box to get the printer name
int GetPrinterName() {
	INT_PTR retDialog;
	extern DWORD Answer;

	retDialog = DialogBoxParam ((HINSTANCE) hModule, 
								MAKEINTRESOURCE(IDD_GetPrinterName), NULL, 
								GetPrinterNameProc, (LPARAM) DEFAULT_PRINTER_NAME);
	
	return(retDialog);
}

/*++

Routine Description:
    Creates the DialogThread, which is responsible for the paper dialog box

Arguments:
    
Return Value:
    TRUE on succesfully creating the thread, else FALSE
--*/

BOOL CreateDialogThread(char *pcText) {

    HANDLE  ThreadHandle;
    DWORD   ThreadId;
 
    //
    // manual-reset event, initially signal state
    //
    
	DlgHwnd = NULL;
	DlgResult = 0;

    ThreadHandle = CreateThread(NULL, 16*1024,
                                (LPTHREAD_START_ROUTINE)DialogThread,
                                pcText, 0, &ThreadId);
    if (ThreadHandle) {
		CloseHandle(ThreadHandle);
        return (TRUE);
    }
	else {
		DlgResult = DLG_ERROR;
		return (FALSE);
	}
}
/*++

Routine Description:
    Print the desired message in a dialog box

Arguments:
	bNewThread: TRUE: Print message in another thread
	PtrOutMessage: <>NULL: Put message to be printed in this string
	dwError:	> 0: The last error code of the callinf thread
					 The error text of this error code is added to the message
				0:	 No error message
	PtrMessage: The message text
	...:	  Additional parameter, which have to be inserted into the message text
    
Return Value:
    TRUE on succesfully creating the thread, else FALSE
--*/

void PrintMessage(BOOL bNewThread, char *PtrOutMessage, DWORD dwError, char *PtrMessage, ...) {
	char	Message[ERROR_STRING_SIZE + 1];
	char	Message1[ERROR_STRING_SIZE + 1];
	char	*cpMessage = Message;
	char	*cpMessage1;
	LPVOID	lpMsgBuf;
	DWORD	dwNumberOfBytes;
	INT_PTR retDialog;
	
	va_list ArgPtr;

	va_start(ArgPtr, PtrMessage);
							// Add the additional parameter to the message
    vsprintf(Message, PtrMessage, ArgPtr);

	if (dwError & USER_ERROR) {
		// special error codes from HPRDH.dll
		switch (dwError) {

			case PRT_OPEN_NOT_DONE:
				cpMessage = Message1;
				sprintf(cpMessage, "0x%x: %s: PrtOpen() not executed", dwError,
						Message);
				break;

			case PRT_CLAIM_NOT_DONE:
				cpMessage = Message1;
				sprintf(cpMessage, "0x%x: %s: PrtClaim() not executed", 
						dwError, Message);
				break;

			case PRT_COMMUNICATION_BROKEN:
				cpMessage = Message1;
				sprintf(cpMessage, "0x%x: %s: \nCommunication with printer broken\n(printer offline?)", 
						dwError, Message);
				break;
		}
	} else {
		if (dwError != 0) {
			// Get the corresponding text to this error code
			dwNumberOfBytes = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
										    FORMAT_MESSAGE_FROM_SYSTEM | 
											FORMAT_MESSAGE_IGNORE_INSERTS,
											NULL, dwError,
											MAKELANGID(LANG_NEUTRAL, 
											SUBLANG_DEFAULT), 
											(LPTSTR) &lpMsgBuf, 0, NULL);
			if (PtrOutMessage != NULL) {
				cpMessage = PtrOutMessage;
			} else {
				cpMessage = Message1;
			}
			cpMessage1 = (char *) lpMsgBuf;
			if (dwNumberOfBytes >= 2) {
				cpMessage1[dwNumberOfBytes - 2] = '\0';
			}
			sprintf(cpMessage, "%ld: %s: %s", dwError, Message, 
					cpMessage1);
			LocalFree(lpMsgBuf);
		}
	}
				// Only print message box, if no other thread is printing one
	WaitForSingleObject(hDonePrintMessage, INFINITE);

	if (bNewThread == TRUE) {
		CreateDialogThread(cpMessage);
	} else {
				// Print the dialog box with the desired message
		retDialog = DialogBoxParam ((HINSTANCE) hModule, 
									MAKEINTRESOURCE(IDD_TOOLBAR), NULL, 
									ToolDlgProc, (LPARAM) cpMessage);
				// Wake up another thread, if it is waiting in WaitForSingleObject()
		SetEvent(hDonePrintMessage);
	}
	return;
}

/*++

Routine Description:
    Entry point of DemoApplication.exe

Arguments:
		argv[0]: "DemoApplication.exe"
		argv[1]: Printer name, if argc > 1
    
Return Value:
    
--*/

int main(int argc, char* argv[])
{
	int Ret;
	
    if (argc > 2) {			// Only printer name is allowed
        usage();
        return(0);
    }

	hModule = GetModuleHandle (NULL); // Get handle to the own module 
			// Create handle to sequence the message windows of the threads
	hDonePrintMessage = CreateEvent(NULL, FALSE, FALSE, NULL);
			// Set event, that the next dialog box can be printed.
	SetEvent(hDonePrintMessage);

	hReadData = CreateEvent(NULL, FALSE, FALSE, NULL);
	SetEvent(hReadData);

	hWakeUp = CreateEvent(NULL, FALSE, FALSE, NULL);
	SetEvent(hWakeUp);
									// Create handle to wake up the read thread
	hWakeupThread = CreateEvent(NULL, FALSE, FALSE, NULL);
		// Reset event, because the read thread normally should wait the timeout
	ResetEvent(hWakeupThread);		// time before the next PrtRead() is done
					// Default value is, that the read thread doesn't have to end
	StopThread = FALSE;	

	// Get the printer name
	memset(PrinterName, (int) '\0', 257);
	if (argc == 2) {
		strncpy(PrinterName, argv[1], 256);
	} else {
		GetPrinterName();	// Ask the user for the printer name in a dialog box
	}

	// Now do the printing
	do {
		Ret = DoPrinting(PrinterName);
		// If the printer name was invalid, ask the user again.
		if (Ret == ERROR_INVALID_PRINTER_NAME) {
			GetPrinterName();
		}
		// ERROR_NOT_READY is returned, if the spooler closed the corresponding
		// port (for instance, if the configuration of the printer is changed
		// between PrtOpen() and PrtClaim())
		// In that case a PrtClose() has to be called and printing has to be
		// restarted with PrtOpen() ...
	} while (	 (Ret == ERROR_INVALID_PRINTER_NAME)
			  || (Ret == ERROR_NOT_READY) );

	return 0;
}


